"""
All tests related to the authentication mechanism.
"""
import pytest

import cshelve

from helpers import unique_key


@pytest.mark.azure
@pytest.mark.parametrize(
    "config_file",
    [
        "tests/configurations/azure-blob/access-key.ini",
        "tests/configurations/azure-blob/connection-string.ini",
        "tests/configurations/azure-blob/standard.ini",
    ],
)
def test_authentication(config_file):
    """
    Test authentication with password and connection string.
    """
    with cshelve.open(config_file) as db:
        key = unique_key + "test_authentication"
        data = "test_authentication"

        # Write data to the DB.
        db[key] = data

        # Data must be accessible in the DB.
        assert db[key] == data

        # Delete the data from the DB.
        del db[key]

    db.close()


@pytest.mark.azure
def test_authentication_read_only():
    """
    Test the read-only authentication.
    """
    can_write_config_file = "tests/configurations/azure-blob/writeable-anonymous.ini"
    read_only_config_file = "tests/configurations/azure-blob/anonymous.ini"

    key = unique_key + "test_authentication_read_only"
    data = "test_authentication_read_only"

    with cshelve.open(can_write_config_file) as db:
        # Write data to the DB.
        db[key] = data

    # The read-only flag is not mandatory, but the underlying implementation will raise an exception if we try to write.
    with cshelve.open(read_only_config_file) as db:
        # Data must be present in the DB.
        assert db[key] == data

    with cshelve.open(can_write_config_file) as db:
        # Delete the data from the DB.
        del db[key]

    db.close()



def write_same_key(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    db['element'] = i

db.close()
"""


def delete_same_key(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    db['element'] = i
    del db['element']

db.close()
"""


def write_several_keys(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    db['element' + str(i)] = i

db.close()
"""


def delete_several_keys(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    k = 'element' + str(i)
    db[k] = i
    del db[k]

db.close()
"""


def read_same_key(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    db['element'] = 0
    r = db[f'element']

db.close()
"""


def read_several_keys(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    k = 'element' + str(i)
    db[k] = i
    r = db[k]

db.close()
"""


def iterate_several_keys(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    k = 'element' + str(i)

for i in range(10):
    for i in db:
        ...

db.close()
"""


def len_several_keys(db_filename: str):
    return f"""
import cshelve

db = cshelve.open('{db_filename}')

for i in range(10):
    k = 'element' + str(i)

for i in range(10):
    _ = len(db)

db.close()
"""


performance_tests = {
    "write_same_key": write_same_key,
    "write_several_keys": write_several_keys,
    "read_same_key": read_same_key,
    "read_several_keys": read_several_keys,
    "delete_same_key": delete_same_key,
    "delete_several_keys": delete_several_keys,
    "iterate_several_keys": iterate_several_keys,
    "len_several_keys": len_several_keys,
}



"""
In-memory storage implementation. Mainly for testing purposes.
"""
from typing import Dict, Iterator

from .provider_interface import ProviderInterface
from .exceptions import key_access


# Contains in-memory databases persisted between open/close during the same program execution.
DB_PERSISTED = {}


class InMemory(ProviderInterface):
    """
    Implements an in-memory database using a dictionary.
    This is mainly for the package and users tests.
    """

    def __init__(self, logger) -> None:
        super().__init__(logger)
        self.db = {}
        self.persist_key = None

        # Following variables are for testing purposes.
        self._created = False
        self._exists = False
        self._synced = False
        self._logging = None
        self._provider_params = None

    def configure_default(self, config: Dict[str, str]) -> None:
        """
        Configure the InMemory client based on the configuration dictionary.
        """
        # If the persist-key configuration is set, the database will be persisted in memory and reused.
        # This is useful when you open/close multiple times the same database (with the same 'persist-key' value).
        self.persist_key = config.get("persist-key")
        # Simulate whether the database exists or must be created.
        self._exists = config.get("exists", "false").lower() == "true"

        # If defined, retrieve the previous database value.
        if self.persist_key:
            # Save the local database pointer to the persisted database.
            if self.persist_key not in DB_PERSISTED:
                DB_PERSISTED[self.persist_key] = self.db
            else:
                self.db = DB_PERSISTED[self.persist_key]

    def configure_logging(self, config: Dict[str, str]) -> None:
        """
        Configure the logging for the InMemory client based on the configuration dictionary.
        """
        self._logging = config

    def set_provider_params(self, provider_params) -> None:
        """
        This method allows the user to specify custom parameters that can't be included in the config.
        """
        self._provider_params = provider_params

    @key_access(KeyError)
    def get(self, key: bytes) -> bytes:
        """
        Retrieve the value of the specified key.

        Args:
            key (bytes): The key to retrieve the value for.

        Returns:
            bytes: The value associated with the key.
        """
        return self.db[key]

    def close(self) -> None:
        """
        Close the database by setting the internal dictionary to None.
        This ensures an error if the user tries to reuse the object.
        """
        self.db = None

    def sync(self) -> None:
        """
        Sync the database. This is a no-op for the in-memory implementation.
        """
        self._synced = True

    def set(self, key: bytes, value: bytes) -> None:
        """
        Add or update an entry in the database.

        Args:
            key (bytes): The key for the entry.
            value (bytes): The value for the entry.
        """
        self.db[key] = value

    @key_access(KeyError)
    def delete(self, key: bytes) -> None:
        """
        Delete an entry from the database.

        Args:
            key (bytes): The key to delete.
        """
        del self.db[key]

    def contains(self, key: bytes) -> bool:
        """
        Check if the specified key exists in the database.

        Args:
            key (bytes): The key to check.

        Returns:
            bool: True if the key exists, False otherwise.
        """
        return key in self.db

    def iter(self) -> Iterator[bytes]:
        """
        Return an iterator over the keys in the database.

        Returns:
            Iterator[bytes]: An iterator over the keys.
        """
        # Convert in list to avoid RuntimeError: dictionary changed size during iteration
        keys = list(self.db.keys())
        yield from keys

    def len(self) -> int:
        """
        Return the number of objects stored in the database.

        Returns:
            int: The number of objects in the database.
        """
        return len(self.db)

    def exists(self) -> bool:
        """
        Check if the database exists.

        Returns:
            bool: True
        """
        return self._exists

    def create(self) -> None:
        """
        Create the database. This is a no-op for the in-memory implementation.
        """
        self._created = True
        self._exists = True



"""
Azure Blob Storage implementation.

This module provides an implementation of the ProviderInterface using Azure Blob Storage for storage.

This module uses an Azure container to store key/value data in blobs.
It creates a blob for each key/value pair, where the key is the blob name and the value is the blob content.
Operations such as iteration and length are performed using the container API.

# To use this module, you need to install:
# - The Azure SDK for Python: `pip install cshelve[azure-blob]`
# - If you want to use passwordless authentication, you also need to install the Azure CLI: https://docs.microsoft.com/en-us/cli/azure/install-azure-cli
"""
import logging
import functools
import io
import os
from typing import Dict, Iterator, Optional

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.storage.blob import BlobType
except ImportError:
    raise ImportError(
        "The Azure SDK for Python is required to use the Azure Blob Storage implementation. "
        "You can install it with `pip install cshelve[azure-blob]`"
    )

from .provider_interface import ProviderInterface
from .exceptions import (
    AuthTypeError,
    AuthArgumentError,
    ConfigurationError,
    key_access,
)

# Max size of the LRU cache for the blob clients.
# Blob clients are cached to avoid creating a new client for each operation.
LRU_CACHE_MAX_SIZE = 2048

# Logs messages.
NO_HANDLER_PROVIDED = "Logging configuration for Azure SDK is set but no handler is provided, logs will be ignored."


class AzureBlobStorage(ProviderInterface):
    """
    Implement the database based on the Azure Blob Storage technology.
    """

    def __init__(self, logger) -> None:
        super().__init__(logger)
        self.container_name = None

        # Azure Blob Storage clients.
        self._blob_service_client = None
        self._container_client = None

        # Azure Blob Storage client configuration.
        self._client_configuration = {}
        # Azure Credential configuration.
        self._credentials_configuration = {}

        # Cache the blob clients to avoid creating a new client for each operation.
        # As the class is not hashable, we can't use the lru_cache directly on the class method and so we wrap it.
        cache_fct = functools.partial(self._get_client_cache)
        self._get_client = functools.lru_cache(maxsize=LRU_CACHE_MAX_SIZE, typed=False)(
            cache_fct
        )

        # Default configuration parameters.
        self.account_url = None
        self.auth_type = None
        self.environment_key = None

        self._provider_parameters = {}

    def configure_default(self, config: Dict[str, str]) -> None:
        """
        Configure the Azure Blob Storage client based on the configuration file.
        """
        # Retrieve the configuration parameters.
        # The Azure Storage Account URL
        # Ex: https://<account_name>.blob.core.windows.net
        self.account_url = config.get("account_url")
        # The authentication type to use.
        # Can be either 'connection_string' or 'passwordless'.
        self.auth_type = config.get("auth_type")
        # The environment variable key that contains the connection string.
        self.environment_key = config.get("environment_key")
        # The name of the container to use.
        # It can be created if it does not exist depending on the flag parameter.
        self.container_name = config.get("container_name")

        if self.container_name is None:
            raise ConfigurationError("Missing container_name in the configuration file")

    def set_provider_params(self, provider_params):
        """
        This method allows the user to specify custom parameters that can't be included in the config.
        """
        self._provider_parameters = provider_params

    @property
    def blob_service_client(self):
        """
        Create the BlobServiceClient and ContainerClient objects when needed.
        """
        if self._blob_service_client is None:
            self._blob_service_client = self.__create_blob_service(
                self.auth_type, self.account_url, self.environment_key
            )
        return self._blob_service_client

    @property
    def container_client(self):
        """
        Only instantiate the container client when it is needed.
        """
        if self._container_client is None:
            self._container_client = self.blob_service_client.get_container_client(
                self.container_name
            )
        return self._container_client

    def configure_logging(self, config: Dict[str, str]) -> None:
        """
        Configure the logging for the InMemory client based on the configuration dictionary.
        """
        if http := config.get("http"):
            self._client_configuration["logging_enable"] = http.lower() == "true"
        if credentials := config.get("credentials"):
            self._credentials_configuration["logging_enable"] = (
                credentials.lower() == "true"
            )

    # If an `ResourceNotFoundError` is raised by the SDK, it is converted to a `KeyError` to follow the `dbm` behavior based on a custom module error.
    @key_access(ResourceNotFoundError)
    def get(self, key: bytes) -> bytes:
        """
        Retrieve the value of the specified key on the Azure Blob Storage container.
        """
        # Azure Blob Storage must be string and not bytes.
        key = key.decode()
        # Init a stream to store the blob content.
        stream = io.BytesIO()

        # Retrieve the blob client.
        client = self._get_client(key)

        # Download the blob content into the stream then return it.
        # The retry pattern and error handling is done by the Azure SDK.
        client.download_blob().readinto(stream)
        return stream.getvalue()

    def close(self) -> None:
        """
        Close the Azure Blob Storage client.
        """
        self.container_client.close()
        self.blob_service_client.close()

    def sync(self) -> None:
        """
        Sync the Azure Blob Storage client.
        """
        # No sync operation is required for Azure Blob Storage.
        ...

    def set(self, key: bytes, value: bytes):
        """
        Create or update the blob with the specified key and value on the Azure Blob Storage container.
        """
        # Azure Blob Storage must be string and not bytes.
        key = key.decode()

        # Retrieve the blob client.
        client = self._get_client(key)

        # Upload the value to a blob named as the key.
        # The retry pattern and error handling is done by the Azure SDK.
        # The blob is overwritten if it already exists.
        # The BlockBlob type is used to store the value as a block blob.
        client.upload_blob(
            value, blob_type=BlobType.BLOCKBLOB, overwrite=True, length=len(value)
        )

    # If an `ResourceNotFoundError` is raised by the SDK, it is converted to a `KeyError` to follow the `dbm` behavior based on a custom module error.
    @key_access(ResourceNotFoundError)
    def delete(self, key: bytes):
        # Azure Blob Storage must be string and not bytes.
        key = key.decode()

        # Retrieve the blob client.
        client = self._get_client(key)

        # Delete the blob.
        # The retry pattern and error handling is done by the Azure SDK.
        client.delete_blob()

    def contains(self, key: bytes) -> bool:
        """
        Return whether the specified key exists on the Azure Blob Storage container.
        """
        # Azure Blob Storage must be string and not bytes.
        return self._get_client(key.decode()).exists()

    def iter(self) -> Iterator[bytes]:
        """
        Return an iterator over the keys in the Azure Blob Storage container.
        """
        for i in self.container_client.list_blob_names():
            # Azure blob names are strings and not bytes.
            # To respect the Shelf interface, we encode the string to bytes.
            yield i.encode()

    def len(self):
        """
        Return the number of objects stored in the database.
        """
        # The Azure SDK does not provide a method to get the number of blobs in a container.
        # We iterate over the blobs and count them.
        return sum(1 for _ in self.container_client.list_blob_names())

    def exists(self) -> bool:
        """
        Check if the container exists on the Azure Blob Storage account.
        """
        return self.blob_service_client.get_container_client(
            self.container_name
        ).exists()

    def create(self):
        """
        Create the container.
        The container must not exist before calling this method.
        """
        self._container_client = self.blob_service_client.create_container(
            self.container_name
        )

    def _get_client_cache(self, key: bytes):
        """
        Cache the blob clients to avoid creating a new client for each operation.
        Size of this object from getsizeof: 48 bytes
        """
        return self.blob_service_client.get_blob_client(self.container_name, key)

    def __create_blob_service(
        self, auth_type: str, account_url: Optional[str], environment_key: Optional[str]
    ):
        # BlobServiceClient and DefaultAzureCredential are imported here to avoid importing them in the module scope.
        # This also simplify the mocking of the Azure SDK in the tests even if it remove the typing information.
        # https://learn.microsoft.com/en-us/python/api/overview/azure/storage-blob-readme?view=azure-python#types-of-credentials
        from azure.storage.blob import BlobServiceClient
        from azure.identity import DefaultAzureCredential

        # Concat users defined parameters and client configuration.
        parameters = {**self._provider_parameters, **self._client_configuration}

        # Create the BlobServiceClient based on the authentication type.
        # A lambda is used to avoid calling the method if the auth_type is not valid.
        supported_auth = {
            "access_key": lambda: BlobServiceClient(
                account_url,
                **parameters,
                credential=self.__get_credentials(environment_key),
            ),
            "anonymous": lambda: BlobServiceClient(account_url, **parameters),
            "connection_string": lambda: BlobServiceClient.from_connection_string(
                self.__get_credentials(environment_key), **parameters
            ),
            # Passwordless authentication is only available with the Azure CLI.
            "passwordless": lambda: BlobServiceClient(
                account_url,
                **parameters,
                credential=DefaultAzureCredential(**self._credentials_configuration),
            ),
        }

        if auth_method := supported_auth.get(auth_type):
            return auth_method()

        raise AuthTypeError(
            f"Invalid auth_type: {auth_type}. Supported values are: {', '.join(supported_auth.keys())}"
        )

    def __get_credentials(self, environment_key: str) -> str:
        """
        Retrieve the credentials from the environment variable or raise the corresponding error.
        """
        if environment_key is None:
            raise AuthArgumentError(f"Missing environment_key parameter")

        if credential := os.environ.get(environment_key):
            return credential

        raise AuthArgumentError(f"Missing environment variable: {environment_key}")



"""
The `open` function accept a `protocol` argument but the `cshelve` module do not interract with it.
The impact is on the algorithm used by the `Shelf` object and not directly on `cshelve`.
Anyway, the library must support it.
"""
import pickle
from unittest.mock import Mock

import cshelve
from cshelve._parser import Config


def test_use_protocol():
    """
    Provide the protocol argument and check if it is passed to the shelve module.
    """
    filename = "test.ini"
    provider = "myprovider"
    protocol = pickle.HIGHEST_PROTOCOL
    config = {
        "provider": provider,
        "auth_type": "passwordless",
        "container_name": "mycontainer",
    }

    cdit = Mock()
    factory = Mock()
    loader = Mock()

    factory.return_value = cdit
    loader.return_value = Config(provider, config, {}, {})

    # Replace the default parser with the mock parser.
    db = cshelve.open(
        filename, protocol=protocol, config_loader=loader, factory=factory
    )

    assert db._protocol == protocol



"""
Verify error handling in the library.
"""
import os
import pytest

import cshelve


CONFIG_FILES = [
    "tests/configurations/azure-blob/standard.ini",
    "tests/configurations/in-memory/persisted.ini",
]


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_key_not_found(config_file):
    """
    Ensure KeyError is raised when key is not found.
    """
    db = cshelve.open(config_file)

    with pytest.raises(cshelve.KeyNotFoundError):
        db["test_key_not_found"]

    db.close()


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_raise_delete_missing_object(config_file):
    """
    Ensure delete an non-existing object raises KeyError.
    """
    db = cshelve.open(config_file)

    key_pattern = "test_delete_object"

    with pytest.raises(cshelve.KeyNotFoundError):
        del db[key_pattern]

    db.close()


def test_unknown_auth_type():
    """
    Ensure exception is raised when auth type is unknown.
    """
    with pytest.raises(cshelve.AuthTypeError):
        cshelve.open(
            "tests/configurations/azure-blob/error-handling/unknown-auth-type.ini"
        )


def test_no_connection_string_key_auth_type():
    """
    Ensure exception is raised when auth type is unknown.
    """
    with pytest.raises(cshelve.AuthArgumentError):
        cshelve.open(
            "tests/configurations/azure-blob/error-handling/connection-string-without-connection-string.ini"
        )


def test_no_connection_string_in_env():
    """
    Ensure exception is raised when auth type is unknown.
    """
    with pytest.raises(cshelve.AuthArgumentError):
        cshelve.open(
            "tests/configurations/azure-blob/error-handling/connection-string-without-env-var.ini"
        )



"""
This Interface defines the interface for storage provider supporting the `MutableMapping` interface.
This class is used by the `Shelf` class to interact with the cloud storage provider.
"""
from abc import abstractmethod
from typing import Dict, Iterator


__all__ = ["ProviderInterface"]


class ProviderInterface:
    """
    This class defines the interface for storage provider to be used by `cshelve`.
    Some methods may be left empty if not needed by the storage provider.
    """

    def __init__(self, logger) -> None:
        self.logger = logger

    @abstractmethod
    def close(self) -> None:
        """
        Close the cloud storage provider.
        """
        raise NotImplementedError

    @abstractmethod
    def configure_default(self, config: Dict[str, str]) -> None:
        """
        Default configuration of the provider.
        """
        raise NotImplementedError

    @abstractmethod
    def configure_logging(self, config: Dict[str, str]) -> None:
        """
        Logging configuration of the provider.
        """
        raise NotImplementedError

    @abstractmethod
    def set_provider_params(self, provider_params) -> None:
        """
        This method allows the user to specify custom parameters that can't be included in the config.
        """
        raise NotImplementedError

    @abstractmethod
    def contains(self, key: bytes) -> bool:
        """
        Check if the key exists.
        """
        raise NotImplementedError

    @abstractmethod
    def create(self) -> None:
        """
        Create the cloud storage provider.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, key: bytes) -> None:
        """
        Delete the key and its associated value.
        """
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        """
        Check if the cloud storage provider exists.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, key: bytes) -> bytes:
        """
        Get the value associated with the key.
        """
        raise NotImplementedError

    @abstractmethod
    def iter(self) -> Iterator[bytes]:
        """
        Return an iterator over the keys.
        """
        raise NotImplementedError

    @abstractmethod
    def len(self) -> int:
        """
        Return the number of keys.
        """
        raise NotImplementedError

    @abstractmethod
    def set(self, key: bytes, value: bytes) -> None:
        """
        Set the value associated with the key.
        """
        raise NotImplementedError

    @abstractmethod
    def sync(self) -> None:
        """
        Sync the cloud storage provider.
        """
        raise NotImplementedError



"""
This module contains the exceptions raised by the cshelve package.
`dbm` exceptions are based on the sub implementations and are not following a standard.
Consequently, we are creating custom exceptions to handle the errors.
"""
from typing import Type


class UnknownProviderError(RuntimeError):
    """
    Raised when an unknown cloud provider is specified in the configuration.
    """

    pass


class UnknownCompressionAlgorithmError(RuntimeError):
    """
    Raised when the compression algorithm provided is incorrect.
    """

    pass


class UnknownEncryptionAlgorithmError(RuntimeError):
    """
    Raised when the encryption algorithm provided is incorrect.
    """

    pass


class KeyNotFoundError(KeyError):
    """
    Raised when a resource is not found.
    """

    pass


class ReadOnlyError(Exception):
    """
    Raised when an attempt is made to write to a read-only DB.
    """

    pass


class DBDoesNotExistsError(Exception):
    """
    Raised when an the DB does not exist and the flag does not allow creating it.
    """

    pass


class CanNotCreateDBError(Exception):
    """
    Raised when an attempt is made to create a DB and it fails.
    """

    pass


class AuthError(Exception):
    """
    Based class for Auth exception.
    """

    pass


class AuthTypeError(AuthError):
    """
    Raised when the auth type is unknown.
    """

    pass


class AuthArgumentError(AuthError):
    """
    Raised when the auth type is unknown.
    """

    pass


class ConfigurationError(RuntimeError):
    """
    Raised when the configuration provided for a provider is incorrect.
    """

    pass


def key_access(exception: Type[Exception]) -> KeyNotFoundError:
    """
    Create a KeyNotFoundError exception if the key is not found.
    """

    def wrapper(func):
        def inner(self, key, *args, **kwargs):
            try:
                return func(self, key, *args, **kwargs)
            except exception as e:
                raise KeyNotFoundError(f"Key not found: {key}") from e

        return inner

    return wrapper



""""
This application is a simple database for the Asterix and Obelix comic series. It allows you to add new users to the database and dump the current users in the database.

How to use:
    python main.py <database-configuration> dump|save|clear
"""
import sys

import cshelve


# Open the database configuration.
with cshelve.open(sys.argv[1], writeback=True) as db:
    # Create the 'users' key if it doesn't exist.
    if "users" not in db:
        db["users"] = []

    # Check the command.
    if sys.argv[2] == "dump":
        # Print all the users in the database.
        for user in db["users"]:
            print(f"{user['name']} {user['insertion-number']}")
    elif sys.argv[2] == "save":
        # Add new users to the database.
        try:
            while user := input():
                user = user.strip()
                db["users"].append(
                    {"name": user, "insertion-number": len(db["users"]) + 1}
                )
            db.sync()
        except EOFError:
            # Catch the EOFError exception when the user presses Ctrl+D.
            ...
    elif sys.argv[2] == "clear":
        # Clear the database.
        db["users"] = []
    else:
        print("Par toutatis, I don't know that command!")



"""
Depending on the filename, either the native shelve module or the cloud shelve module is used.
The cloud shelve module is used when the filename has a specific extension, and we must ensure that the correct module is used.
"""
from pathlib import Path
from unittest.mock import Mock
from cshelve._parser import load, use_local_shelf


def test_use_local_shelf():
    """
    If the filename is not finishing by '.ini', the default shelve module must be used.
    """
    fallback_default_module = [Path("test.sqlite3"), Path("test.db"), Path("test.dat")]

    for filename in fallback_default_module:
        assert use_local_shelf(filename) is True
        # assert use_local_shelf(Path(filename)) is True


def test_use_cloud_shelf():
    """
    If the filename is finishing by '.ini', the cloud shelve module must be used.
    """
    cloud_module = [Path("test.ini"), Path("cloud.ini"), Path("test.cloud.ini")]

    for filename in cloud_module:
        assert use_local_shelf(filename) is False


def test_azure_configuration():
    """
    Load the Azure configuration file and return it as a dictionary.
    """
    config = load(Mock(), Path("tests/configurations/azure-blob/standard.ini"))

    assert config.provider == "azure-blob"

    assert config.default["auth_type"] == "connection_string"
    assert config.default["environment_key"] == "AZURE_STORAGE_CONNECTION_STRING"
    assert config.default["container_name"] == "standard"

    assert config.logging["http"] == "true"
    assert config.logging["credentials"] == "false"
    assert config.logging["level"] == "INFO"



"""
Based on the [official doc](https://docs.python.org/3/library/dbm.html#dbm.open), flags must be one of the following:
- 'r' (default): Open existing database for reading only.
- 'w': Open existing database for reading and writing.
- 'c': Open database for reading and writing, creating it if it doesn't exist.
- 'n': Always create a new, empty database, open for reading and writing.

The module must adapt its behavior based on the flag provided by the user and restrict or process some operations based on the flag.
"""
import functools

from .exceptions import ReadOnlyError


def can_create(flag: str) -> bool:
    """
    Returns True if a new database can be created.
    """
    return flag in ("c", "n")


def can_write(func) -> bool:
    """
    This decorator restricts write operations based on the flag.
    If the operation is not allowed, it raises a `ReadOnlyError`.
    """

    @functools.wraps(func)
    def can_write(obj, *args, **kwargs):
        if obj.flag == "r":
            raise ReadOnlyError("Reader can't store")
        return func(obj, *args, **kwargs)

    return can_write


def clear_db(flag: str) -> bool:
    """
    Returns True if the user requests to clear the database.
    """
    return flag == "n"



"""
Ensure the standard behavior of the API works as expected in real scenarios.
"""
import os

import pytest

import cshelve

from helpers import write_data, unique_key, del_data


CONFIG_FILES = [
    "tests/configurations/azure-blob/standard.ini",
    "tests/configurations/in-memory/persisted.ini",
    "tests/configurations/azure-blob/compression.ini",
    "tests/configurations/in-memory/compression.ini",
]

CONFIG_FILES_ITER = [
    "tests/configurations/azure-blob/iter.ini",
    "tests/configurations/in-memory/iter.ini",
]

CONFIG_FILES_LEN = [
    "tests/configurations/azure-blob/len.ini",
    "tests/configurations/in-memory/len.ini",
]

CONFIG_FILES_DEL = [
    "tests/configurations/azure-blob/del.ini",
    "tests/configurations/in-memory/del.ini",
]

CONFIG_FILES_FLAG_N = [
    "tests/configurations/azure-blob/flag-n.ini",
    "tests/configurations/in-memory/flag-n.ini",
]


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_write_then_read(config_file: str):
    """
    Ensure we can read and write data to the DB.
    """
    with cshelve.open(config_file) as db:
        key_pattern = unique_key + "test_write_and_read"
        data_pattern = "test_write_and_read"

        for i in range(10):
            key = f"{key_pattern}{i}"

            # Write data to the DB.
            db[key] = f"{data_pattern}{i}"
            # Data must be present in the DB.
            assert db[key] == f"{data_pattern}{i}"
            # Delete the data from the DB.
            del db[key]

    db.close()


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_read_after_reopening(config_file: str):
    """
    Ensure the data is still present after reopening the DB.
    """
    key_pattern = unique_key + "test_read_after_reopening"
    data_pattern = "test_read_after_reopening"

    def read_data():
        db = cshelve.open(config_file)

        for i in range(10):
            key = f"{key_pattern}{i}"
            assert db[key] == f"{data_pattern}{i}"
            del db[key]

        db.close()

    write_data(config_file, key_pattern, data_pattern)
    read_data()


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_update_on_operator(config_file: str):
    """
    Ensure operator interface works as expected.
    """
    key_pattern = unique_key + "test_update_on_operator"
    str_data_pattern = "test_update_on_operator"
    list_data_pattern = [1]

    def write_data():
        db = cshelve.open(config_file)

        for i in range(10):
            db[f"{key_pattern}{i}"] = str_data_pattern
            db[f"{key_pattern}{i}-list"] = list_data_pattern

        db.close()

    def update_data():
        db = cshelve.open(config_file)

        for i in range(10):
            db[f"{key_pattern}{i}"] += f"{i}"
            db[f"{key_pattern}{i}-list"] += [i]

        db.close()

    def read_data():
        db = cshelve.open(config_file)

        for i in range(10):
            key = f"{key_pattern}{i}"
            key_list = f"{key_pattern}{i}-list"

            # Operator `+=` on string does not modify the original string.
            assert db[key] == f"{str_data_pattern}{i}"
            # Operator `+=` on list does modify the original list.
            assert db[key_list] == list_data_pattern + [i]

            del db[key]
            del db[key_list]

        db.close()

    write_data()
    update_data()
    read_data()


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_contains(config_file: str):
    """
    Ensure __contains__ works as expected.
    """
    db = cshelve.open(config_file)

    key_pattern = unique_key + "test_contains"
    data_pattern = "test_contains"

    db[key_pattern] = data_pattern

    assert key_pattern in db

    del db[key_pattern]


@pytest.mark.parametrize("config_file", CONFIG_FILES_FLAG_N)
def test_clear_db(config_file):
    """
    Ensure the database is cleared when using the 'n' flag.
    """
    key_pattern = "test_clear_db"
    data_pattern = "test_clear_db"

    def rewrite_db():
        db = cshelve.open(config_file, "n")

        assert len(db) == 0

        for i in range(10):
            db[f"{key_pattern}{i}"] = f"{data_pattern}{i}"

        db.close()

    def read_data():
        db = cshelve.open(config_file, "r")

        for i in range(10):
            key = f"{key_pattern}{i}"
            assert db[key] == f"{data_pattern}{i}"

        db.close()

    write_data(config_file, key_pattern, data_pattern)
    rewrite_db()
    read_data()
    del_data(config_file, key_pattern)


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES_DEL,
)
def test_del(config_file):
    """
    Ensure we can delete a record from the DB.
    """
    key_pattern = "test_del"
    data_pattern = "test_del"

    def _del_data():
        db = cshelve.open(config_file)

        for i in range(10):
            key = f"{key_pattern}{i}"
            assert db[key] == f"{data_pattern}{i}"
            del db[key]

        assert len(db) == 0
        db.close()

    write_data(config_file, key_pattern, data_pattern)
    _del_data()


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES_LEN,
)
def test_len(config_file):
    """
    Ensure __len__ works as expected.
    """
    db = cshelve.open(config_file)

    key_pattern = "test_len"
    data_pattern = "test_len"

    del_data(config_file)

    for i in range(10):
        db[f"{key_pattern}{i}"] = f"{data_pattern}{i}"

    assert len(db) == 10

    for i in range(10):
        del db[f"{key_pattern}{i}"]

    assert len(db) == 0


@pytest.mark.parametrize("config_file", CONFIG_FILES_ITER)
def test_iter(config_file):
    """
    Ensure the __iter__ method works as expected.
    """
    res = set()
    db = cshelve.open(config_file)

    key_pattern = "test_iter"
    data_pattern = "test_iter"
    del_data(config_file)

    for i in range(10):
        key = f"{key_pattern}{i}"
        db[key] = f"{data_pattern}{i}"
        res.add(key)

    keys = set(db)
    assert keys == res

    db.close()
    del_data(config_file)



"""
The factory ensures that the correct backend is loaded based on the provider.
"""
from unittest.mock import patch, Mock

import pytest

from cshelve import UnknownProviderError
from cshelve._factory import factory


@patch("cshelve._azure_blob_storage.AzureBlobStorage")
def test_known_backend(azure_mock):
    """
    Test factory loading the Azure backend.
    """
    factory(Mock(), "azure-blob")

    azure_mock.assert_called_once()


def test_unknown_backend():
    """
    Ensure that the factory raises an error when an unknown backend is requested.
    """
    with pytest.raises(UnknownProviderError):
        factory(Mock(), "aws")



"""
This example demonstrates the flexibility of the database by dealing with types and entries around the theme of Asterix and Obelix.
Here is the initial structure of the database:
{
    'gaule': ['Adrénaline', 'Amérix', ...],
    'rome': ['Anglaigus', 'Caius Obtus', ...],
    'new_friends': {}
}
And here is the structure of the database after adding 'Cléopâtre' and 'Numérobis' from 'Egypte' and 'Chipolata' from 'Corse':
{
    'gaule': ['Adrénaline', 'Amérix', ...],
    'rome': ['Anglaigus', 'Caius Obtus', ...],
    'new_friends': {
        'cléopâtre': 'egypte',
        'numérobis': 'egypte',
        'chipolata': 'corse'
    },
    'egypte': ['cléopâtre', 'numérobis'],
    'corse': ['chipolata']
}

How to use it:
1. Install the necessary package: pip install -r requirements.txt
2. Run the program: python3 main.py mydb
"""
import cshelve
import sys

# Some Gaulois and Romans are added at the startup.
GAULOIS_KEY = "gaule"
ROMANS_KEY = "rome"
# This key will contain new friends added to simplify dictionary operations.
NEW_FRIENDS_KEY = "new_friends"

# Open the database based on the program parameter.
db = cshelve.open(sys.argv[1], "n")

# Load the Gaulois.
if GAULOIS_KEY not in db:
    with open("./gaulois.txt", "r") as fd:
        # Lowercase the names to simplify the comparison.
        # Remove empty strings.
        db[GAULOIS_KEY] = [p.lower() for p in fd.read().split("\n") if p]

# Load the Romans.
if ROMANS_KEY not in db:
    with open("./romans.txt", "r") as fd:
        # Lowercase the names to simplify the comparison.
        # Remove empty strings.
        db[ROMANS_KEY] = [p.lower() for p in fd.read().split("\n") if p]

# Add the default key to simplify dictionary parsing.
if NEW_FRIENDS_KEY not in db:
    db[NEW_FRIENDS_KEY] = {}

print("Enter 'end' to finish our adventure.")

while True:
    # Remove potential '\n' and unnecessary spaces.
    personnage = input("Enter the name of the character: ").strip().lower()

    if not personnage:
        continue

    # If the character is a Gaulois, print a custom message.
    if personnage in db[GAULOIS_KEY]:
        print("A true Gaulois par Toutatis!")
    # If the character is a Roman, print a custom message.
    elif personnage in db[ROMANS_KEY]:
        print("A Roman!")
    # End of the loop.
    elif personnage == "end":
        break
    # This character was added by the user.
    # Let's retrieve where he/she is from and display it.
    elif _from := db[NEW_FRIENDS_KEY].get(personnage):
        print(f"{personnage} is a true {_from}")
    # Unknown character, let's add it to the DB.
    else:
        _from = (
            input(f"I don't know {personnage}, where is he/she from? ").strip().lower()
        )
        # Unknown location.
        if _from not in db:
            db[_from] = [personnage]
        # Some characters are already from this location, we want to keep them too.
        else:
            db[_from] += [personnage]
        print("Thanks for this new friend!")
        # Add the new friend to the nested dict to simplify the parsing.
        db[NEW_FRIENDS_KEY] = {**db[NEW_FRIENDS_KEY], personnage: _from}

# Display friends and where they are from.
for _from, personages in db.items():
    if NEW_FRIENDS_KEY == _from:
        continue
    print(f"Here are friend(s) from {_from.capitalize()}:")
    for p in personages:
        print(f"- {p.capitalize()}")

db.close()



"""
Compression module for cshelve.
"""
from functools import partial
from logging import Logger
from typing import Dict

from ._data_processing import DataProcessing
from .exceptions import UnknownCompressionAlgorithmError


ALGORITHMS_NAME_KEY = "algorithm"
COMPRESSION_LEVEL_KEY = "level"


def configure(
    logger: Logger, data_processing: DataProcessing, config: Dict[str, str]
) -> None:
    """
    Configure the compression algorithm.
    """
    # Compression is not configured, silently return.
    if not config:
        return

    if ALGORITHMS_NAME_KEY not in config:
        logger.info("No compression algorithm specified.")
        return

    algorithm = config[ALGORITHMS_NAME_KEY]

    supported_algorithms = {
        "zlib": _zlib,
    }

    if compression := supported_algorithms.get(algorithm):
        logger.debug(f"Configuring compression algorithm: {algorithm}")
        compression_fct, decompression_fct = compression(config)
        data_processing.add_pre_processing(compression_fct)
        data_processing.add_post_processing(decompression_fct)
        logger.debug(f"Compression algorithm {algorithm} configured.")
    else:
        raise UnknownCompressionAlgorithmError(
            f"Unsupported compression algorithm: {algorithm}"
        )


def _zlib(config: Dict[str, str]):
    """
    Configure zlib compression.
    """
    import zlib

    level = int(config.get(COMPRESSION_LEVEL_KEY, zlib.Z_DEFAULT_COMPRESSION))

    compress = partial(zlib.compress, level=level)
    decompress = partial(zlib.decompress)

    return compress, decompress



import sys
import timeit
import cshelve
import datetime
from concurrent.futures import ThreadPoolExecutor

from tests import performance_tests


BACKENDS = [
    # Run test locally.
    "local",
    # Run test using Azure Blob Storage.
    "test-azure.ini",
]

# The database containing the result of the tests.
database_name = sys.argv[1]

# Retrieve the OS type, the Python version and the commit hash.
# Those information allows to track performance on different platforms.
# If not provided an default version is used to allow test.
if len(sys.argv) > 2:
    OS_TYPE = sys.argv[2]
    PYTHON_MAJOR_VERSION = sys.argv[3]
    COMMIT_HASH = sys.argv[4]
else:
    OS_TYPE = "unknown"
    PYTHON_MAJOR_VERSION = "unknown"
    COMMIT_HASH = "unknown"


def save(db, backend_name, fct_name, exec_time):
    """
    Save the result of a performance test in the database.

    Data is stored in the following format:
    {
        '<backend_name>': {
            '<fct_name>': [
                {
                  'exec_time': <exec_time>,
                  'datetime': <datetime>,
                  'commit_hash': '<commit_hash>'
                }
                ...
            ],
            ...
        },
        ...
    }
    """
    # Add the backend in the DB if not exists.
    if backend_name not in db:
        db[backend_name] = {}

    # Retrieve backend data.
    backend_perfs = db[backend_name]
    # Add the res to the backend data
    if fct_name not in backend_perfs:
        backend_perfs[fct_name] = []

    backend_perfs[fct_name].append(
        {
            "exec_time": exec_time,
            "datetime": datetime.datetime.now(),
            "commit_hash": COMMIT_HASH,
        }
    )

    # Save the result in the DB.
    db[backend_name] = backend_perfs


def run_test(db, backend):
    res_key = get_key(backend)

    for name, fct in performance_tests.items():
        print(
            f"Running test {name} for backend {backend} on {OS_TYPE} with Python {PYTHON_MAJOR_VERSION}.",
            flush=True,
        )
        # Purge the DB used by tests.
        cshelve.open(backend, "n").close()
        # Execute the test providing the backend to test.
        res_test_perf = fct(backend)
        exec_time = timeit.timeit(res_test_perf, number=10)
        # Save the result in the DB.
        save(db, res_key, name, exec_time)


def get_key(backend):
    return f"{backend}--{OS_TYPE}--{PYTHON_MAJOR_VERSION}"


with cshelve.open(database_name) as db:
    # Run the tests for each backend and save the results in the result DB.
    # Because backend are independent, we can run them in parallel.
    # It may have an impact on the network, but currently the number of backend is limited.
    with ThreadPoolExecutor(max_workers=len(BACKENDS)) as executor:
        list(executor.map(lambda backend: run_test(db, backend), BACKENDS))

    # Simpli display the results.
    print("Results:")
    backend_names = [get_key(b) for b in BACKENDS]
    for backend_name, res in db.items():
        if backend_name not in backend_names:
            continue
        for fct_name, res in res.items():
            if len(res) > 1:
                previous = res[-2]["exec_time"] if len(res) > 1 else 0
                current = res[-1]["exec_time"]
                perf_result = (
                    f"better: -({previous - current:.2f})"
                    if current < previous
                    else f"worst (+{current - previous:.2f})"
                )
                print(
                    f"Backend: {backend_name}, Function: {fct_name}, diff performance are {perf_result}"
                )
            else:
                print(
                    f"No previous performance to compare with for backend {backend_name}, Function {fct_name}."
                )



"""
Encryption module for cshelve.
"""
from functools import partial
from logging import Logger
from typing import Dict

from ._data_processing import DataProcessing
from .exceptions import UnknownEncryptionAlgorithmError


ALGORITHMS_NAME_KEY = "algorithm"
COMPRESSION_LEVEL_KEY = "level"


def configure(
    logger: Logger, data_processing: DataProcessing, config: Dict[str, str]
) -> None:
    """
    Configure the encryption algorithm.
    """
    # Encryption is not configured, silently return.
    if not config:
        return

    if ALGORITHMS_NAME_KEY not in config:
        logger.info("No compression algorithm specified.")
        return

    algorithm = config[ALGORITHMS_NAME_KEY]

    supported_algorithms = {
        "aes256": _aes256,
    }

    if encryption := supported_algorithms.get(algorithm):
        logger.debug(f"Configuring encryption algorithm: {algorithm}")
        crypt_fct, decrypt_fct = encryption(config)
        data_processing.add_pre_processing(crypt_fct)
        data_processing.add_post_processing(decrypt_fct)
        logger.debug(f"Encryption algorithm {algorithm} configured.")
    else:
        raise UnknownEncryptionAlgorithmError(
            f"Unsupported encryption algorithm: {algorithm}"
        )


def _aes256(config: Dict[str, str]):
    """
    Configure aes256 encryption.
    """
    import zlib

    crypt = lambda x: x
    decrypt = lambda x: x

    return crypt, decrypt



"""
Helper functions for the integration tests.
"""
import sys
from typing import Optional
import cshelve


# When the CI is running, tests are executed on multiple platforms and Python versions accessing the same container with the same key.
# To avoid conflicts, we use a unique key for each platform and Python version.
unique_key = sys.platform + str(sys.version_info.minor)


def write_data(db_config: str, key_pattern: str, data_pattern: str):
    """
    Write data to the DB.
    """
    db = cshelve.open(db_config)

    for i in range(10):
        db[f"{key_pattern}{i}"] = f"{data_pattern}{i}"

    db.close()


def del_data(config_file: str, key_pattern: Optional[str] = None):
    """
    Delete data from the DB.
    """
    db = cshelve.open(config_file)

    if key_pattern is None:
        for i in db:
            del db[i]
    else:
        for i in range(10):
            del db[f"{key_pattern}{i}"]

    db.close()



"""
This module contains the integration tests for the writeback functionality.
The `Shelf` object manages this feature but the `cshelve` module must provide all the necessary tools.
"""
import os
import pytest
import cshelve

from helpers import unique_key

CONFIG_FILES = [
    "tests/configurations/azure-blob/standard.ini",
    "tests/configurations/in-memory/persisted.ini",
]


@pytest.mark.parametrize("config_file", CONFIG_FILES)
def test_writeback(config_file):
    """
    Ensure the writeback functionality works as expected.
    """
    key_pattern = unique_key + "test_writeback"
    data_pattern = [1]

    def _write_data():
        db = cshelve.open(config_file)

        for i in range(10):
            db[f"{key_pattern}{i}"] = data_pattern

        db.close()

    def update_data(writeback):
        db = cshelve.open(config_file, writeback=writeback)

        for i in range(10):
            key = f"{key_pattern}{i}"
            value = db[key]
            value.append(i)

            if writeback:
                assert db[key] == data_pattern + [i]
            else:
                assert db[key] == data_pattern

        db.close()

    def read_data(contains_index):
        db = cshelve.open(config_file)

        for i in range(10):
            key = f"{key_pattern}{i}"
            if contains_index:
                assert db[key] == data_pattern + [i]
            else:
                assert db[key] == data_pattern

        db.close()

    def del_data():
        db = cshelve.open(config_file)

        for i in range(10):
            del db[f"{key_pattern}{i}"]

        db.close()

    # Write default data.
    _write_data()
    # Update data with writeback=False, so data must not be updated.
    update_data(writeback=False)
    # Ensure data was not updated.
    read_data(contains_index=False)
    # Update data with writeback=True, so data must be updated.
    update_data(writeback=True)
    # Ensure data was updated.
    read_data(contains_index=True)
    del_data()



# """
# Test the encryption module.
# """
# from unittest.mock import Mock

# import pytest

# from cshelve import UnknownEncryptionAlgorithmError
# from cshelve._encryption import configure
# from cshelve._data_processing import DataProcessing


# @pytest.fixture
# def data_processing():
#     return DataProcessing()


# def test_no_encryption(data_processing):
#     """
#     Ensure nothing si configure when the config is empty.
#     """
#     logger = Mock()
#     config = {}

#     configure(logger, data_processing, config)

#     assert len(data_processing.post_processing) == 0
#     assert len(data_processing.pre_processing) == 0


# def test_default_aes256_config(data_processing):
#     """
#     Ensure AES256 is configured when defined.
#     """
#     logger = Mock()
#     config = {"algorithm": "aes256"}

#     configure(logger, data_processing, config)

#     assert len(data_processing.post_processing) == 1
#     assert len(data_processing.pre_processing) == 1
#     assert data_processing.pre_processing[0].func == None
#     assert data_processing.post_processing[0].func == None

#     first_pre_processing_applied = id(data_processing.pre_processing[0])
#     first_post_processing_applied = id(data_processing.post_processing[0])

#     # Ensure the same behaviours and order if configured twice.
#     configure(logger, data_processing, config)

#     assert len(data_processing.post_processing) == 2
#     assert len(data_processing.pre_processing) == 2
#     # Ensure the order is respected.
#     assert first_pre_processing_applied == id(data_processing.pre_processing[0])
#     assert first_post_processing_applied == id(data_processing.post_processing[0])


# def test_unknowned_algorithm(data_processing):
#     """
#     Ensure an exception is raised when an unknowed algorithm is provided.
#     """
#     logger = Mock()
#     config = {"algorithm": "unknow"}

#     with pytest.raises(UnknownEncryptionAlgorithmError):
#         configure(logger, data_processing, config)



"""
Package entry point exposing the `open` function to open a cloud shelf and exceptions.

The `open` function is the main entry point of the package.
Based on the file extension, it will open a local or cloud shelf, but in any case, it will return a `shelve.Shelf` object.

If the file extension is `.ini`, the file is considered a configuration file and handled by `cshelve`; otherwise, it will be handled by the standard `shelve` module.
"""
import logging
from pathlib import Path
import shelve

from ._data_processing import DataProcessing
from ._database import _Database
from ._compression import configure as _configure_compression
from ._factory import factory as _factory
from ._parser import load as _config_loader
from ._parser import use_local_shelf
from .exceptions import (
    AuthArgumentError,
    AuthTypeError,
    CanNotCreateDBError,
    ConfigurationError,
    DBDoesNotExistsError,
    KeyNotFoundError,
    ReadOnlyError,
    UnknownCompressionAlgorithmError,
    UnknownEncryptionAlgorithmError,
    UnknownProviderError,
)


__all__ = [
    "AuthArgumentError",
    "AuthTypeError",
    "CanNotCreateDBError",
    "ConfigurationError",
    "DataProcessing",
    "DBDoesNotExistsError",
    "KeyNotFoundError",
    "open",
    "ReadOnlyError",
    "ResourceNotFoundError",
    "UnknownCompressionAlgorithmError",
    "UnknownProviderError",
]


class CloudShelf(shelve.Shelf):
    """
    A cloud shelf is a shelf that is stored in the cloud. It is a subclass of `shelve.Shelf` and is used to store data in the cloud.

    The underlying storage provider is provided by the factory based on the provider name then abstract by the _Database facade.
    """

    def __init__(
        self,
        filename,
        flag,
        protocol,
        writeback,
        config_loader,
        factory,
        logger,
        provider_params,
    ):
        # Load the configuration file to retrieve the provider and its configuration.
        config = config_loader(logger, filename)

        # Let the factory create the provider interface object based on the provider name then configure it.
        provider_interface = factory(logger, config.provider)
        provider_interface.configure_logging(config.logging)
        provider_interface.configure_default(config.default)
        provider_interface.set_provider_params(provider_params)

        # Data processing object used to apply pre and post processing to the data.
        data_processing = DataProcessing()
        _configure_compression(logger, data_processing, config.compression)

        # The CloudDatabase object is the class that interacts with the cloud storage backend.
        # This class doesn't perform or respect the shelve.Shelf logic and interface so we need to wrap it.
        database = _Database(logger, provider_interface, flag, data_processing)
        database._init()

        # Let the standard shelve.Shelf class handle the rest.
        super().__init__(database, protocol, writeback)


def open(
    filename,
    flag="c",
    protocol=None,
    writeback=False,
    config_loader=_config_loader,
    factory=_factory,
    logger=logging.getLogger("cshelve"),
    provider_params={},
) -> shelve.Shelf:
    """
    Open a cloud shelf or a local shelf based on the file extension.
    """
    # Ensure the filename is a Path object.
    filename = Path(filename)

    if use_local_shelf(filename):
        logger.debug("Opening a local shelf.")
        # The user requests a local and not a cloud shelf.
        # Dependending of the Python version, the shelve module doesn't accept Path objects.
        return shelve.open(str(filename), flag, protocol, writeback)

    logger.debug("Opening a cloud shelf.")
    return CloudShelf(
        filename,
        flag.lower(),
        protocol,
        writeback,
        config_loader,
        factory,
        logger,
        provider_params,
    )



"""
Depending on the filename, either the native shelve module or the cloud shelve module is used.
The cloud shelve module is used when the filename has a specific extension, and we must ensure that the correct module is used.
"""
from pathlib import Path
import shelve
import tempfile
from unittest.mock import Mock

import cshelve
from cshelve._parser import Config


def test_load_cloud_shelf_config():
    """
    Based on the filename, the cloud shelve module must be used.
    At the same time, we test the parser injection functionality.
    """
    filename = "test.ini"
    provider = "myprovider"
    default_config = {
        "provider": provider,
        "auth_type": "passwordless",
        "container_name": "mycontainer",
    }
    logging_config = {"http": "true", "credentials": "false"}
    compression_config = {}

    cloud_database = Mock()
    factory = Mock()
    loader = Mock()
    logger = Mock()
    attended_filename = Path(filename)

    factory.return_value = cloud_database
    loader.return_value = Config(
        provider, default_config, logging_config, compression_config
    )
    cloud_database.exists.return_value = False

    # Replace the default parser with the mock parser.
    with cshelve.open(
        filename, config_loader=loader, factory=factory, logger=logger
    ) as cs:
        loader.assert_called_once_with(logger, attended_filename)
        factory.assert_called_once_with(logger, provider)
        assert isinstance(cs.dict.db, Mock)
        cs.dict.db.configure_default.assert_called_once_with(default_config)
        cs.dict.db.configure_logging.assert_called_once_with(logging_config)


def test_load_cloud_shelf_config_memory():
    """
    Test the open function and the configuration.
    """
    filename = "tests/configurations/in-memory/not-persisted.ini"

    # Replace the default parser with the mock parser.
    with cshelve.open(filename) as cs:
        # Ensure the default configuration is loaded.
        cs.persist_key = False
        # Ensure the logging is loaded.
        cs._config = {"enabled": "true"}
        # Ensure the database is created.
        cs._created = True


def test_load_local_shelf_config():
    """
    Based on the filename, the default shelve module must be used.
    """
    local_shelf_suffix = [".sqlite3", ".db", ".dat"]

    for suffix in local_shelf_suffix:
        # When instanciate, shelf modules create the file with the provided name.
        # So we create a temporary file to garbage collect it after the test.
        with tempfile.NamedTemporaryFile(suffix=suffix) as fp:
            fp.close()
            with cshelve.open(fp.name) as db:
                assert isinstance(db, shelve.DbfilenameShelf)


def test_support_pathlib():
    """
    Provide a pathlib object to the open function and ensure the correct filename is used.
    """
    filepath = Path("tests/configurations/in-memory/persisted.ini")

    with cshelve.open(filepath) as cs:
        cs["key"] = "value"

    with cshelve.open(filepath) as cs:
        assert cs["key"] == "value"


def test_parameters():
    """
    Ensure users parameters are provided to the provider.
    """
    filepath = Path("tests/configurations/in-memory/persisted.ini")
    params = {"foo": "bar", "number": 42}

    with cshelve.open(filepath, provider_params=params) as cs:
        cs._user_parameters = params



"""
Factory module to return the correct module to be used.
"""
from logging import Logger
from .provider_interface import ProviderInterface
from .exceptions import UnknownProviderError


def factory(logger: Logger, provider: str) -> ProviderInterface:
    """
    Return the correct module to be used.
    """
    logger.debug(f"Creating the provider '{provider}'...")
    res = _factory(logger, provider)
    logger.debug("Provider created.")
    return res


def _factory(logger: Logger, provider: str):
    logger.info(f"Loading provider {provider}")

    if provider == "azure-blob":
        from ._azure_blob_storage import AzureBlobStorage

        return AzureBlobStorage(logger)
    elif provider == "in-memory":
        from ._in_memory import InMemory

        return InMemory(logger)

    logger.critical("Provider not found.")
    raise UnknownProviderError(f"Provider Interface '{provider}' is not supported.")



from unittest.mock import Mock

import pytest
from cshelve._data_processing import DataProcessing
from cshelve._database import _Database
from cshelve._in_memory import InMemory
from cshelve.exceptions import CanNotCreateDBError, DBDoesNotExistsError, ReadOnlyError


@pytest.fixture
def database() -> _Database:
    flag = "c"
    logger = Mock()
    provider_db = InMemory(logger)
    data_processing = DataProcessing()
    db = _Database(logger, provider_db, flag, data_processing)
    db._init()
    return db


def test_setitem(database):
    """
    Ensure that the __setitem__ method returns the value associated with the key from the database.
    """
    key, value = b"key", b"value"
    database[key] = value

    assert value == database.db.get(key)


def test_getitem(database):
    """
    Ensure that the __getitem__ method returns the value associated with the key from the database.
    """
    key, value = b"key", b"value"
    database.db.set(key, value)

    assert value == database[key]


def test_delitem(database):
    """
    Ensure that the __delitem__ method deletes the key from the database.
    """
    key, value = b"key", b"value"
    database.db.set(key, value)

    del database[key]

    assert False == database.db.contains(key)


def test_iter(database):
    """
    Ensure that the __iter__ method iterates over the keys in the database.
    """
    key, value = b"key", b"value"
    key2, value2 = b"key2", b"value2"

    database.db.set(key, value)
    database.db.set(key2, value2)

    assert list(database) == [key, key2]


def test_len(database):
    """
    Ensure that the __len__ method returns the number of elements in the database.
    """
    database.db.set("key", "value")
    database.db.set("key2", "value2")

    assert len(database) == 2


def test_close(database):
    """
    Ensure that the close method closes the database.
    """
    database.close()

    assert database.db.db is None


def test_sync(database):
    """
    Ensure that the sync method syncs the database.
    """
    assert database.db._synced == False

    database.sync()

    assert database.db._synced == True


def test_doesnt_create_database_if_exists():
    """
    Ensure the database is not created if it already exists.
    """
    logger = Mock()
    flag = "c"
    provider_db = InMemory(logger)
    provider_db.configure_default({"exists": "True"})

    assert provider_db._created == False

    db = _Database(logger, provider_db, flag, DataProcessing())
    db._init()

    assert provider_db._created == False


def test_create_database_if_not_exists():
    """
    Ensure the database is created if it doesn't exist.
    """
    logger = Mock()
    flag = "c"
    provider_db = InMemory(logger)

    assert provider_db._created == False
    assert provider_db._exists == False

    db = _Database(logger, provider_db, flag, DataProcessing())
    db._init()

    assert provider_db._created == True


def test_cant_create_database_if_not_exists_and_not_allowed():
    """
    Ensure exception is raised if the database doesn't exist and the flag doesn't allow it.
    """
    logger = Mock()
    flags = "r", "w"

    for flag in flags:
        provider_db = InMemory(logger)

        assert provider_db._created == False
        assert provider_db._exists == False

        db = _Database(logger, provider_db, flag, DataProcessing())

        with pytest.raises(DBDoesNotExistsError) as _:
            db._init()


def test_error_database_creation():
    """
    Ensure an internal exception is raised if the database can't be created.
    """
    logger = Mock()
    provider_db = Mock()
    flag = "c"

    provider_db.exists.return_value = False
    provider_db.create.side_effect = Exception
    db = _Database(logger, provider_db, flag, DataProcessing())

    with pytest.raises(CanNotCreateDBError) as _:
        db._init()


def test_database_clear_if_asked():
    """
    Ensure the database is cleared if the flag allows it.
    """
    logger = Mock()
    flag = "n"
    provider_db = InMemory(logger)
    provider_db.configure_default({"exists": "True"})

    provider_db.set("key", "value")
    provider_db.set("key2", "value2")

    assert provider_db.len() == 2
    db = _Database(logger, provider_db, flag, DataProcessing())
    db._init()
    assert provider_db.len() == 0


def test_do_not_clear_database():
    """
    Ensure the database is not cleared if the flag doesn't allow it.
    """
    logger = Mock()
    flags = "r", "w", "c"

    for flag in flags:
        provider_db = InMemory(logger)
        provider_db.configure_default({"exists": "True"})

        provider_db.set("key", "value")
        provider_db.set("key2", "value2")

        assert provider_db.len() == 2
        db = _Database(logger, provider_db, flag, DataProcessing())
        db._init()
        assert provider_db.len() == 2


def test_read_only():
    """
    Ensure the database is not cleared if the flag doesn't allow it.
    """
    logger = Mock()
    flag = "r"
    key, value = b"key", b"value"
    new_key, new_value = b"key-new", b"value-new"

    provider_db = InMemory(logger)
    provider_db.configure_default({"exists": "True"})

    provider_db.set(key, value)

    db = _Database(logger, provider_db, flag, DataProcessing())
    db._init()

    with pytest.raises(ReadOnlyError) as _:
        db[new_key] = new_value

    with pytest.raises(ReadOnlyError) as _:
        del db[key]

    assert db[key] == value



"""
This file contains the integration tests for the open flag.
The `n` flag is tested in the [sequential tests](../sequential/test_sequential.py) due to its global impact.
"""
import os
import pytest

import cshelve

from helpers import write_data, del_data, unique_key


CONFIG_FILES = [
    "tests/configurations/azure-blob/flag.ini",
    "tests/configurations/in-memory/persisted.ini",
]


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_read_only(config_file):
    """
    A read-only database should not allow writing and must raise an exception if we try to do so.
    The exception is raised by the implementation of the `dbm` module and not by `shelve` itself, so a custom exception is raised.
    """
    key_pattern = unique_key + "test_read_only"
    data_pattern = "test_read_only"

    def cant_update():
        db = cshelve.open(config_file, "r")

        for i in range(10):
            key = f"{key_pattern}{i}"

            assert db[key] == f"{data_pattern}{i}"

            with pytest.raises(cshelve.ReadOnlyError):
                db[key] = data_pattern

        db.close()

    write_data(config_file, key_pattern, data_pattern)
    cant_update()
    del_data(config_file, key_pattern)


@pytest.mark.azure
def test_container_does_not_exists():
    """
    Depending of the flag, the database must already exists otherwise an exception is raised.
    The exception is raised by the implementation of the `dbm` module and not by `shelve` itself, so a custom exception is raised.
    """
    with pytest.raises(cshelve.DBDoesNotExistsError):
        cshelve.open(
            "tests/configurations/azure-blob/container-does-not-exists.ini",
            "w",
        )

    with pytest.raises(cshelve.DBDoesNotExistsError):
        cshelve.open(
            "tests/configurations/azure-blob/container-does-not-exists.ini",
            "r",
        )



"""
Integration tests for the protocol parameter.
The `Shelf` object mainly manages this functionality, but we must ensure `cshelve` can handle it.
"""
import os
import pickle
import pytest

import cshelve

from helpers import unique_key


CONFIG_FILES = [
    "tests/configurations/azure-blob/standard.ini",
    "tests/configurations/in-memory/persisted.ini",
]


@pytest.mark.parametrize("config_file", CONFIG_FILES)
def test_protocol(config_file):
    """
    Ensure cshelve works correctly with the non default protocol.
    """
    key_pattern = unique_key + "test_protocol"
    data_pattern = "test_protocol"
    protocol = pickle.HIGHEST_PROTOCOL

    with cshelve.open(config_file, protocol=protocol) as db:
        for i in range(10):
            db[f"{key_pattern}{i}"] = f"{data_pattern}{i}"

    with cshelve.open(config_file, protocol=protocol) as db:
        for i in range(10):
            assert f"{data_pattern}{i}" == db[f"{key_pattern}{i}"]


@pytest.mark.parametrize("config_file", CONFIG_FILES)
def test_change_protocol(config_file):
    """
    Ensure cshelve works correctly with the non default protocol.
    """
    key_pattern = unique_key + "test_protocol"
    data_pattern = "test_protocol"
    protocol = pickle.HIGHEST_PROTOCOL

    with cshelve.open(config_file, protocol=protocol) as db:
        for i in range(10):
            db[f"{key_pattern}{i}"] = f"{data_pattern}{i}"

    with cshelve.open(config_file, protocol=protocol) as db:
        for i in range(10):
            assert f"{data_pattern}{i}" == db[f"{key_pattern}{i}"]



"""
This module provides the DataProcessing class, which handles pre-processing and post-processing of data.

Examples:
    >>> dp = DataProcessing()
    >>> dp.add_pre_processing(lambda x: x + b'1')
    >>> dp.add_post_processing(lambda x: x + b'2')
    >>> pre_processed = dp.apply_pre_processing(b'0')
    >>> pre_processed
    b'01'
    >>> post_processed = dp.apply_post_processing(pre_processed)
    >>> post_processed
    b'012'
"""
from typing import Callable, Optional, List


class DataProcessing:
    """
    A class to handle pre-processing and post-processing of data.
    """

    def __init__(
        self,
        pre_processing: Optional[List[Callable[[bytes], bytes]]] = None,
        post_processing: Optional[List[Callable[[bytes], bytes]]] = None,
    ):
        """
        Initializes the DataProcessing class with optional pre-processing and post-processing lists.

        Examples:
        >>> dp = DataProcessing()
        >>> dp.pre_processing
        []
        >>> dp.post_processing
        []
        """
        self.pre_processing = pre_processing if pre_processing is not None else []
        self.post_processing = post_processing if post_processing is not None else []

    def add_pre_processing(self, func: Callable[[bytes], bytes]) -> None:
        """
        Adds a function to the pre-processing list.

        Examples:
        >>> dp = DataProcessing()
        >>> dp.add_pre_processing(lambda x: x + 1)
        >>> len(dp.pre_processing)
        1
        """
        self.pre_processing.append(func)

    def add_post_processing(self, func: Callable[[bytes], bytes]) -> None:
        """
        Adds a function to the post-processing list.

        Args:
        func (function): A function to add to the post-processing list.

        Examples:
        >>> dp = DataProcessing()
        >>> dp.add_post_processing(lambda x: x * 2)
        >>> len(dp.post_processing)
        1
        """
        self.post_processing.append(func)

    def apply_pre_processing(self, data: bytes) -> bytes:
        """
        Applies all pre-processing functions to the data.

        Args:
        data: The data to process.

        Returns:
        The processed data.

        Examples:
        >>> dp = DataProcessing(pre_processing=[lambda x: x + 1, lambda x: x * 2])
        >>> dp.apply_pre_processing(1)
        4
        """
        for func in self.pre_processing:
            data = func(data)
        return data

    def apply_post_processing(self, data: bytes) -> bytes:
        """
        Applies all post-processing functions to the data.

        Args:
        data: The data to process.

        Returns:
        The processed data.

        Examples:
        >>> dp = DataProcessing(post_processing=[lambda x: x / 2, lambda x: x - 1])
        >>> dp.apply_post_processing(4)
        1.0
        """
        for func in self.post_processing:
            data = func(data)
        return data



"""
MutableMapping interface for the database.

The term _Database is used to follow the naming convention of the Python Shelve module, even though it is not mandatory.
"""
from logging import Logger
from collections.abc import MutableMapping
from concurrent.futures import ThreadPoolExecutor

from ._data_processing import DataProcessing
from .provider_interface import ProviderInterface
from ._flag import can_create, can_write, clear_db
from .exceptions import (
    CanNotCreateDBError,
    DBDoesNotExistsError,
    DBDoesNotExistsError,
)


__all__ = ["_Database"]


class _Database(MutableMapping):
    """
    Wrapper around the ProviderInterface to provide a MutableMapping interface with the Shelf business logic.
    """

    def __init__(
        self,
        logger: Logger,
        db: ProviderInterface,
        flag: str,
        data_processing: DataProcessing,
    ) -> None:
        super().__init__()
        self.data_processing = data_processing
        self.db = db
        self.flag = flag
        self.logger = logger

    def __getitem__(self, key: bytes) -> bytes:
        """
        Retrieve the value associated with the key from the database.
        """
        value = self.db.get(key)
        return self.data_processing.apply_post_processing(value)

    @can_write
    def __setitem__(self, key: bytes, value: bytes) -> None:
        """
        Set the value associated with the key in the database.
        """
        value = self.data_processing.apply_pre_processing(value)
        self.db.set(key, value)

    @can_write
    def __delitem__(self, key: bytes) -> None:
        """
        Delete the key from the database.
        """
        self.db.delete(key)

    def __iter__(self):
        """
        Iterate over the keys in the database.
        """
        yield from self.db.iter()

    def __len__(self) -> int:
        """
        Return the number of elements in the database.
        """
        return self.db.len()

    def close(self) -> None:
        """
        Close the database.
        """
        self.db.close()

    def sync(self) -> None:
        """
        Sync the database.
        """
        self.db.sync()

    def _init(self):
        """
        Initialize the database by:
        - Creating the database if it doesn't exist and the flag allows it.
        - Clearing the database if the flag allows it.
        """
        if not self.db.exists():
            self.logger.info(f"Database doesn't exists.")
            if can_create(self.flag):
                self.logger.info(f"Creating the database...")
                try:
                    self.db.create()
                except Exception as e:
                    self.logger.critical(f"Can't create the database.")
                    raise CanNotCreateDBError("Can't create database.") from e
                self.logger.info(f"Database created.")
            else:
                self.logger.critical(f"Can't create the database")
                raise DBDoesNotExistsError("Database does not exist.")
        else:
            # If the database exists, but the flag parameter indicates that it should be cleared, clear it.
            if clear_db(self.flag):
                self.logger.info(f"Purging the database...")
                # Retrieve all the keys and delete them.
                # Retrieving keys is quick, but deletion synchronously is slow, so we use threads to speed up the process.
                with ThreadPoolExecutor() as executor:
                    list(executor.map(self.db.delete, self.db.iter()))
                self.logger.info(f"Database purged.")



"""
The factory ensures that the correct backend is loaded based on the provider.
"""
import pickle
from unittest.mock import Mock

from cshelve import CloudShelf
from cshelve._parser import Config


def test_factory_usage():
    """
    End users may want to provide another factory to create the cloud database and not the default one.
    This test ensures that the factory provided is used.
    """
    filename = "does_not_exists.ini"
    provider = "fake"
    config = {42: 42}
    compression = {}
    flag = "c"
    protocol = pickle.HIGHEST_PROTOCOL
    writeback = False

    cloud_database = Mock()
    factory = Mock()
    loader = Mock()
    logger = Mock()

    loader.return_value = Config(provider, config, config, compression)
    factory.return_value = cloud_database
    cloud_database.exists.return_value = False

    with CloudShelf(
        filename,
        flag,
        protocol,
        writeback,
        config_loader=loader,
        factory=factory,
        logger=logger,
        provider_params={},
    ) as cs:
        cloud_database.exists.assert_called_once()
        factory.assert_called_once_with(logger, provider)
        # The mock returned by the factory must be the MuttableMapping object used by the shelve.Shelf object.
        assert isinstance(cs.dict.db, Mock)


def test_loader_usage():
    """
    Users may want to provide another loader to load the configuration file.
    This test ensures that the loader provided is used.
    """
    filename = "does_not_exists.ini"
    provider = "fake"
    config = {42: 42}
    compression = {}
    flag = "c"
    protocol = pickle.HIGHEST_PROTOCOL
    writeback = False

    cloud_database = Mock()
    factory = Mock()
    loader = Mock()
    logger = Mock()

    loader.return_value = Config(provider, config, config, compression)
    factory.return_value = cloud_database
    cloud_database.exists.return_value = False

    with CloudShelf(
        filename,
        flag,
        protocol,
        writeback,
        config_loader=loader,
        factory=factory,
        logger=logger,
        provider_params={},
    ) as cs:
        loader.assert_called_once_with(logger, filename)



"""
Test the compression module.
"""
from unittest.mock import Mock
import zlib

import pytest

from cshelve import UnknownCompressionAlgorithmError
from cshelve._compression import configure
from cshelve._data_processing import DataProcessing


@pytest.fixture
def data_processing():
    return DataProcessing()


def test_no_compression(data_processing):
    """
    Ensure nothing si configure when the config is empty.
    """
    logger = Mock()
    config = {}

    configure(logger, data_processing, config)

    assert len(data_processing.post_processing) == 0
    assert len(data_processing.pre_processing) == 0


def test_default_zlib_config(data_processing):
    """
    Ensure Zlib is configured when defined.
    If no level is provided, the default compression must be set.
    """
    logger = Mock()
    config = {"algorithm": "zlib"}

    configure(logger, data_processing, config)

    assert len(data_processing.post_processing) == 1
    assert len(data_processing.pre_processing) == 1
    assert data_processing.pre_processing[0].func == zlib.compress
    assert data_processing.post_processing[0].func == zlib.decompress
    assert (
        data_processing.pre_processing[0].keywords["level"]
        == zlib.Z_DEFAULT_COMPRESSION
    )
    assert data_processing.post_processing[0].keywords == {}

    first_pre_processing_applied = id(data_processing.pre_processing[0])
    first_post_processing_applied = id(data_processing.post_processing[0])

    # Ensure the same behaviours and order if configured twice.
    configure(logger, data_processing, config)

    assert len(data_processing.post_processing) == 2
    assert len(data_processing.pre_processing) == 2
    # Ensure the order is respected.
    assert first_pre_processing_applied == id(data_processing.pre_processing[0])
    assert first_post_processing_applied == id(data_processing.post_processing[0])


def test_zlib_level(data_processing):
    """
    Ensure the Zlib configuration level can be configured.
    """
    logger = Mock()
    compression_level = 5
    config = {"algorithm": "zlib", "level": compression_level}

    configure(logger, data_processing, config)

    assert data_processing.pre_processing[0].keywords["level"] == compression_level


def test_unknowned_algorithm(data_processing):
    """
    Ensure an exception is raised when an unknowed algorithm is provided.
    """
    logger = Mock()
    config = {"algorithm": "unknow"}

    with pytest.raises(UnknownCompressionAlgorithmError):
        configure(logger, data_processing, config)



import logging
from unittest.mock import patch, Mock
import pytest
from azure.storage.blob import BlobType

from cshelve._factory import factory
from azure.core.exceptions import ResourceNotFoundError
from cshelve import (
    AuthArgumentError,
    AuthTypeError,
    KeyNotFoundError,
    ConfigurationError,
)


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_passwordless(BlobServiceClient, DefaultAzureCredential):
    """
    Test the Azure Blob Storage client with the passwordless authentication.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": "container",
    }
    identity = Mock()
    DefaultAzureCredential.return_value = identity

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.create()

    DefaultAzureCredential.assert_called_once()
    BlobServiceClient.assert_called_once_with(
        config["account_url"],
        credential=identity,
    )


@patch("azure.storage.blob.BlobServiceClient")
def test_anonymous_public_read_access(BlobServiceClient):
    """
    Ensure the capability of accessing an Azure Blob Storage client with the anonymous public read access authentication.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "anonymous",
        "container_name": "container",
    }

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.create()

    BlobServiceClient.assert_called_once_with(config["account_url"])


@patch("azure.storage.blob.BlobServiceClient")
def test_connection_string(BlobServiceClient):
    """
    Test the Azure Blob Storage client with the connection string authentication.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "connection_string",
        "environment_key": "ENV_VAR",
        "container_name": "container",
    }
    connection_string = "my_connection_string"

    with patch.dict("os.environ", {"ENV_VAR": connection_string}):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)
        provider.create()

    BlobServiceClient.from_connection_string.assert_called_once_with(connection_string)


@patch("azure.storage.blob.BlobServiceClient")
def test_account_key(BlobServiceClient):
    """
    Test the Azure Blob Storage client with an account key authentication.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "access_key",
        "environment_key": "ENV_VAR",
        "container_name": "container",
    }
    access_key = "my_access_key"

    with patch.dict("os.environ", {"ENV_VAR": access_key}):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)
        provider.create()

    BlobServiceClient.assert_called_once_with(
        config["account_url"],
        credential=access_key,
    )


@pytest.mark.parametrize(
    "auth_type",
    ["access_key", "connection_string"],
)
def test_missing_env_var(auth_type):
    """
    Test the Azure Blob Storage client with the connection string authentication without the env variable
    key in the configuration.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": auth_type,
        "container_name": "container",
    }

    with pytest.raises(AuthArgumentError):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)
        provider.create()


@pytest.mark.parametrize(
    "auth_type",
    ["access_key", "connection_string"],
)
def test_missing_container(auth_type):
    """
    Ensure a ConfigurationError is raised when the container name is missing from the configuration.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": auth_type,
    }

    with pytest.raises(ConfigurationError):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)


@pytest.mark.parametrize(
    "auth_type",
    ["access_key", "connection_string"],
)
def test_missing_env_var_value(auth_type):
    """
    Test the Azure Blob Storage client with the connection string/account key authentication without providing the
    connection string in the environment variable.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": auth_type,
        "environment_key": "ENV_VAR",
        "container_name": "container",
    }

    with pytest.raises(AuthArgumentError):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)
        provider.create()


def test_wrong_auth_type():
    """
    Ensure an AuthTypeError is raised when the authentication type is not supported.
    """
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "unknonwn",
        "container_name": "container",
    }

    with pytest.raises(AuthTypeError):
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config)
        provider.create()


@patch("io.BytesIO")
@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_get(BlobServiceClient, DefaultAzureCredential, BytesIO):
    """
    Ensure we can retrieve a value from an Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    key = b"key"

    blob_client = Mock()
    blob_service_client = Mock()
    download_blob = Mock()
    stream = Mock()

    BytesIO.return_value = stream
    DefaultAzureCredential.return_value = Mock()
    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_blob_client.return_value = blob_client
    blob_client.download_blob.return_value = download_blob

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    provider.get(key)

    # Blob is retrieved from the container.
    blob_service_client.get_blob_client.assert_called_once_with(
        container_name, key.decode()
    )
    # Create the blob content stream.
    blob_client.download_blob.assert_called_once()
    # Blob content is read into the stream.
    download_blob.readinto.assert_called_once()
    # Stream content is returned.
    stream.getvalue.assert_called_once()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_get_key_error(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure a key error is raised when the blob is not found in the Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    key = b"doesnt-exists"

    blob_client = Mock()
    blob_service_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_blob_client.return_value = blob_client
    blob_client.download_blob.side_effect = ResourceNotFoundError

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    with pytest.raises(KeyNotFoundError):
        provider.get(key)


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_set(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure we can set a value to a Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    key, value = b"key", b"value"

    blob_client = Mock()
    blob_service_client = Mock()
    upload_blob = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_blob_client.return_value = blob_client
    blob_client.upload_blob.return_value = upload_blob

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    provider.set(key, value)

    # Blob is retrieved from the container.
    blob_service_client.get_blob_client.assert_called_once_with(
        container_name, key.decode()
    )
    # Upload the blob content.
    blob_client.upload_blob.assert_called_once_with(
        value,
        blob_type=BlobType.BLOCKBLOB,
        overwrite=True,
        length=len(value),
    )


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_close(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the close method is called on the Azure Blob Storage client.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }

    blob_service_client = Mock()
    container_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_container_client.return_value = container_client

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.close()

    container_client.close.assert_called_once()
    blob_service_client.close.assert_called_once()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_delete(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the delete method is called on the Azure Blob Storage client.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    key = b"key"

    blob_client = Mock()
    blob_service_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_blob_client.return_value = blob_client

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    provider.delete(key)

    # Blob is retrieved from the container.
    blob_service_client.get_blob_client.assert_called_once_with(
        container_name, key.decode()
    )
    # Blob is deleted.
    blob_client.delete_blob.assert_called_once()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_iter(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the list of key is correctly returned from the Azure Blob Storage during iteration.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    list_blob_names = ["key1", "key2"]
    list_blob_names_attended = [b"key1", b"key2"]

    blob_service_client = Mock()
    container_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_container_client.return_value = container_client
    container_client.list_blob_names.return_value = list_blob_names

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    assert list(provider.iter()) == list_blob_names_attended


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_contains(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the contains correctly check if the blob exists on the Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    key = b"key"

    blob_client = Mock()
    blob_service_client = Mock()
    upload_blob = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_blob_client.return_value = blob_client

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    provider.contains(key)

    # Blob is retrieved from the container.
    blob_service_client.get_blob_client.assert_called_once_with(
        container_name, key.decode()
    )
    # Ensure the exists method is called.
    blob_client.exists.assert_called_once()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_len(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the cacul of the number of keys is correctly returned from the Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    list_blob_names = ["key1", "key2"]

    blob_service_client = Mock()
    container_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_container_client.return_value = container_client
    container_client.list_blob_names.return_value = list_blob_names

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)

    assert 2 == provider.len()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_exists(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the exists method is correctly called from the Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }

    blob_service_client = Mock()
    container_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client
    blob_service_client.get_container_client.return_value = container_client

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.exists()

    container_client.exists.assert_called_once()


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_create(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the create method on a container is correctly called from the Azure Blob Storage.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }

    blob_service_client = Mock()
    DefaultAzureCredential.return_value = Mock()

    BlobServiceClient.return_value = blob_service_client

    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.create()

    blob_service_client.create_container.assert_called_once_with(container_name)


@pytest.mark.parametrize(
    "auth_type",
    ["passwordless", "anonymous", "access_key"],
)
@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_http_logging(BlobServiceClient, DefaultAzureCredential, auth_type):
    """
    Ensure the correct logging configuration is applied to the Azure Blob Storage client.
    """
    container_name = "container"
    config_default = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": auth_type,
        "container_name": container_name,
        "environment_key": "ENV_VAR",
    }
    enable_logging = {"http": "true"}
    disable_logging = {"http": "false"}

    with patch.dict("os.environ", {"ENV_VAR": "my_connection_string"}):
        # Ensure the default behaviour of the azure-blob-storage package.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.create()
        assert "logging_enable" not in BlobServiceClient.call_args.kwargs
        DefaultAzureCredential.reset_mock()

        # Ensure the logging is disabled.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.configure_logging(disable_logging)
        provider.create()
        assert "logging_enable" in BlobServiceClient.call_args.kwargs
        assert BlobServiceClient.call_args.kwargs["logging_enable"] == False
        DefaultAzureCredential.reset_mock()

        # Ensure the logging is enable.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.configure_logging(enable_logging)
        provider.create()
        assert "logging_enable" in BlobServiceClient.call_args.kwargs
        assert BlobServiceClient.call_args.kwargs["logging_enable"] == True


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_http_logging_connection_string(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the correct logging configuration is applied to the Azure Blob Storage client.
    """
    container_name = "container"
    config_default = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "connection_string",
        "container_name": container_name,
        "environment_key": "ENV_VAR",
    }
    enable_logging = {"http": "true"}
    disable_logging = {"http": "false"}

    with patch.dict("os.environ", {"ENV_VAR": "my_connection_string"}):
        # Ensure the default behaviour of the azure-blob-storage package.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.create()
        assert (
            "logging_enable"
            not in BlobServiceClient.from_connection_string.call_args.kwargs
        )
        DefaultAzureCredential.reset_mock()

        # Ensure the logging is disabled.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.configure_logging(disable_logging)
        provider.create()
        assert (
            "logging_enable"
            in BlobServiceClient.from_connection_string.call_args.kwargs
        )
        assert (
            BlobServiceClient.from_connection_string.call_args.kwargs["logging_enable"]
            == False
        )
        DefaultAzureCredential.reset_mock()

        # Ensure the logging is enable.
        provider = factory(Mock(), "azure-blob")
        provider.configure_default(config_default)
        provider.configure_logging(enable_logging)
        provider.create()
        assert (
            "logging_enable"
            in BlobServiceClient.from_connection_string.call_args.kwargs
        )
        assert (
            BlobServiceClient.from_connection_string.call_args.kwargs["logging_enable"]
            == True
        )


@patch("azure.identity.DefaultAzureCredential")
@patch("azure.storage.blob.BlobServiceClient")
def test_credential_logging(BlobServiceClient, DefaultAzureCredential):
    """
    Ensure the correct logging configuration is applied to the Azure Blob Storage client.
    """
    container_name = "container"
    config = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "passwordless",
        "container_name": container_name,
    }
    enable_credentials_logging = {"credentials": "true"}
    disable_credentials_logging = {"credentials": "false"}

    # Ensure the default behaviour of the azure-blob-storage package.
    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.create()
    assert "logging_enable" not in DefaultAzureCredential.call_args.kwargs
    DefaultAzureCredential.reset_mock()

    # Ensure the logging is disabled.
    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.configure_logging(disable_credentials_logging)
    provider.create()
    assert "logging_enable" in DefaultAzureCredential.call_args.kwargs
    assert DefaultAzureCredential.call_args.kwargs["logging_enable"] == False
    DefaultAzureCredential.reset_mock()

    # Ensure the logging is enable.
    provider = factory(Mock(), "azure-blob")
    provider.configure_default(config)
    provider.configure_logging(enable_credentials_logging)
    provider.create()
    assert "logging_enable" in DefaultAzureCredential.call_args.kwargs
    assert DefaultAzureCredential.call_args.kwargs["logging_enable"] == True
    DefaultAzureCredential.reset_mock()


@patch("azure.storage.blob.BlobServiceClient")
def test_provider_params(BlobServiceClient):
    """
    Ensure provider parameters are used.
    """
    config_default = {
        "account_url": "https://account.blob.core.windows.net",
        "auth_type": "anonymous",
        "container_name": "container",
    }
    config_logging = {"http": "true"}
    params = {"secondary_hostname": "second", "max_single_put_size": 42}

    provider = factory(Mock(), "azure-blob")
    provider.configure_logging(config_logging)
    provider.configure_default(config_default)
    provider.set_provider_params(params)
    provider.create()

    BlobServiceClient.assert_called_once_with(
        config_default["account_url"], logging_enable=True, **params
    )



"""
Ensure that the flag parameter is correctly handled and permissions are correctly set.

The [official documentation](https://docs.python.org/3/library/dbm.html#dbm.open) for the flag parameter is as follows:
- 'r' (default): Open existing database for reading only.
- 'w': Open existing database for reading and writing.
- 'c': Open database for reading and writing, creating it if it doesn't exist.
- 'n': Always create a new, empty database, open for reading and writing.
"""
from unittest.mock import Mock

import pytest

import cshelve
from cshelve._flag import can_create, can_write, clear_db


def test_read_only():
    """
    # The `shelve` module raises an exception when trying to write to a read-only database.
    # The exception is raised by the module used by `shelve` and not `shelve` itself, so it is a `cshelve` exception that we have to catch.
    """
    mock = Mock()
    mock.flag = "r"

    with pytest.raises(cshelve.ReadOnlyError):
        can_write(lambda x: x)(mock)


def test_can_write():
    """
    As explain in the official doc, the 'w', 'c', and 'n' flags should allow writing to the database.
    """
    mock = Mock()

    for flag in ("w", "c", "n"):
        mock.flag = flag
        assert can_write(lambda x: True)(mock) == True


def test_can_create():
    """
    As explain in the official doc, the 'c' and 'n' flags should allow creating a new database.
    """
    for flag in ("c", "n"):
        assert can_create(flag) == True


def test_can_not_create():
    """
    Creating a new database is not allowed with the 'w' and 'r' flags.
    """
    for flag in ("w", "r"):
        assert can_create(flag) == False


def test_do_not_clear_db():
    """
    A database open with the 'n' flag empty it before using it.
    """
    assert clear_db("n") == True


def test_do_not_clear_db():
    """
    Enure that the database is not cleared when the 'n' flags is not used.
    """
    for flag in ("w", "r", "c"):
        assert clear_db(flag) == False



"""
This module is responsible for parsing the configuration file.
It reads the configuration file and returns the provider and its configuration as a dictionary.
It also provides a function to determine if a local shelf should be used based on the file extension.

At this level, the only necessary configuration is the provider name.
Other configurations are loaded into a dictionary and passed to the provider for further configuration.
"""
from logging import Logger
from collections import namedtuple
import configparser
from pathlib import Path
from typing import Dict, Tuple


# Default ini section containing the provider and its configuration.
DEFAULT_CONFIG_STORE = "default"
# Key containing the provider name.
PROVIDER_KEY = "provider"
# Logging configuration section.
LOGGING_KEY_STORE = "logging"
# Compression configuration section.
COMPRESSION_KEY_STORE = "compression"


# Tuple containing the provider name and its configuration.
Config = namedtuple("Config", ["provider", "default", "logging", "compression"])


def use_local_shelf(filename: Path) -> bool:
    """
    If the user specify a filename with an extension different of '.ini', a local shelf (the standard library) must be used.
    """
    return not filename.suffix == ".ini"


def load(logger: Logger, filename: Path) -> Tuple[str, Dict[str, str]]:
    """
    Load the configuration file and return it as a dictionary.
    """
    logger.debug(f"Loading configuration file: {filename}.")
    config = configparser.ConfigParser()
    config.read(filename)

    c = config[DEFAULT_CONFIG_STORE]
    logging_config = config[LOGGING_KEY_STORE] if LOGGING_KEY_STORE in config else {}
    compression_config = (
        config[COMPRESSION_KEY_STORE] if COMPRESSION_KEY_STORE in config else {}
    )

    logger.debug(f"Configuration file '{filename}' loaded.")
    return Config(
        provider=c[PROVIDER_KEY],
        default=c,
        logging=logging_config,
        compression=compression_config,
    )



from unittest.mock import Mock
import pytest

from cshelve._factory import factory
from cshelve import KeyNotFoundError


def test_get_and_set():
    """
    Ensure we can set then retrieve a value from the database.
    """
    key, value = b"key", b"value"

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    provider.set(key, value)
    assert value == provider.get(key)


def test_do_not_persist():
    """
    Ensure the DB is by default purged between open operation.
    """
    key, value = b"key", b"value"

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    provider.set(key, value)
    # Ensure the element is persisted.
    assert provider.len() == 1

    # Re open the database to ensure the element is gone.
    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    assert provider.len() == 0


def test_get_and_set_persist():
    """
    Ensure we can set then retrieve a value from the database.
    """
    config = {"persist-key": "only-in-memory"}
    key, value = b"key", b"value"

    provider = factory(Mock(), "in-memory")
    provider.configure_default(config)

    provider.set(key, value)
    # Ensure the element is persisted.
    assert provider.len() == 1

    # Re open the database to ensure the element is still there.
    provider = factory(Mock(), "in-memory")
    provider.configure_default(config)

    assert provider.len() == 1


def test_get_key_error():
    """
    Ensure a key error is raised when the key is not found.
    """
    key = b"does-not-exist"

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    with pytest.raises(KeyNotFoundError):
        provider.get(key)


def test_close():
    """
    Ensure the close method can be called.
    """
    provider = factory(Mock(), "in-memory")
    provider.configure_default({})
    provider.close()


def test_delete():
    """
    Ensure the delete method can be called.
    """
    key, value = b"key", b"value"

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    provider.set(key, value)
    assert provider.contains(key)
    provider.delete(key)
    assert not provider.contains(key)


def test_iter():
    """
    Ensure the list of key is correctly returned.
    """
    keys = [b"key1", b"key2"]

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    for k in keys:
        provider.set(k, b"value")

    assert list(provider.iter()) == keys


def test_contains():
    """
    Ensure contains correctly works.
    """
    key = b"key"

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    assert not provider.contains(key)
    provider.set(key, b"value")
    assert provider.contains(key)


def test_len():
    """
    Ensure the calculation of the number of keys is correctly returned.
    """
    list_blob_names = [b"key1", b"key2"]

    provider = factory(Mock(), "in-memory")
    provider.configure_default({})

    for k in list_blob_names:
        provider.set(k, b"value")

    assert 2 == provider.len()


def test_exists():
    """
    Ensure the exists method can be called.
    """
    provider = factory(Mock(), "in-memory")
    provider.configure_default({})
    provider.exists()


def test_create():
    """
    Ensure the create method can be called.
    """
    provider = factory(Mock(), "in-memory")
    provider.configure_default({})
    provider.create()



"""
This module test data processing modules.
"""
import pickle
import zlib
import cshelve

from helpers import unique_key


def test_compression():
    """
    Ensure the data is compressed.
    """
    compressed_configuration = "tests/configurations/in-memory/compression.ini"
    key_pattern = unique_key + "test_writeback"
    data = "this must be compressed"

    with cshelve.open(compressed_configuration) as db:
        db[key_pattern] = data

        assert (
            pickle.loads(zlib.decompress(db.dict.db.db[key_pattern.encode()])) == data
        )



from cshelve import DataProcessing


def test_add_pre_processing():
    dp = DataProcessing()
    dp.add_pre_processing(lambda x: x + 1)
    assert len(dp.pre_processing) == 1
    assert dp.pre_processing[0](1) == 2


def test_add_post_processing():
    dp = DataProcessing()
    dp.add_post_processing(lambda x: x * 2)
    assert len(dp.post_processing) == 1
    assert dp.post_processing[0](2) == 4


def test_apply_pre_processing():
    dp = DataProcessing(pre_processing=[lambda x: x + 1, lambda x: x * 2])
    result = dp.apply_pre_processing(1)
    assert result == 4  # (1 + 1) * 2


def test_apply_post_processing():
    dp = DataProcessing(post_processing=[lambda x: x / 2, lambda x: x - 1])
    result = dp.apply_post_processing(4)
    assert result == 1  # (4 / 2) - 1


def test_full_processing():
    dp = DataProcessing(
        pre_processing=[lambda x: x + 1, lambda x: x * 2],
        post_processing=[lambda x: x / 2, lambda x: x - 1],
    )

    data = 1
    pre_processed_data = dp.apply_pre_processing(data)
    final_data = dp.apply_post_processing(pre_processed_data)

    assert pre_processed_data == 4  # (1 + 1) * 2
    assert final_data == 1  # (4 / 2) - 1



"""
Ensure the library can handle large data.
"""
import os
import pytest
import numpy as np
import pandas as pd

import cshelve

from helpers import unique_key

CONFIG_FILES = [
    "tests/configurations/azure-blob/standard.ini",
    "tests/configurations/in-memory/persisted.ini",
]


@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
)
def test_large(config_file):
    """
    Update a relative large DataFrame in the DB to verify it is possible.
    """
    db = cshelve.open(config_file)

    key_pattern = unique_key + "test_large"

    # 167.46 MiB
    df = pd.DataFrame(
        np.random.randint(0, 100, size=(844221, 26)),
        columns=list("ABCDEFGHIGKLMNOPQRSTUVWXYZ"),
    )

    db[key_pattern] = df
    new_df = db[key_pattern]

    assert id(new_df) != id(df)
    assert new_df.equals(df)
